@@ -2,7 +2,6 @@ import fastremap
 import numpy as np
 import tqdm
 import imageio
-from annotationframeworkclient.frameworkclient import FrameworkClient
 import cloudvolume as cv
 from scipy import ndimage
 from functools import partial
@@ -13,36 +12,30 @@ import imageio
 from . import utils
 
 
-def _patch_cloudvolume_auth(token):
-    # This is a a hack at the moment, but we will replace the monkey patching when cloudvolume can properly accept tokens.
-    global cv
-    cv.datasource.graphene.metadata.chunkedgraph_credentials = {'token': token}
-
-
-def bounds_from_center(ctr, delx=0, dely=0, delz=0):
-    """Generate bounds from a center point and halfwidths in each direction
+def bounds_from_center(ctr, width=1, height=1, depth=1):
+    """Generate bounds from a center point and dimensions for each direction
 
     Parameters
     ----------
     ctr : array-like
         x,y,z coordinates of the center of the bounds in voxel dimensions.
-    delx : int, optional 
-        Distance from the center to the edge of the box in the x direction.
-        Default is 0.
-    dely : int, optional 
-        Distance from the center to the edge of the box in the y direction.
-        Default is 0.
-    delz : int, optional 
-        Distance from the center to the edge of the box in the z direction.
-        Default is 0.
+    width: int, optional
+        Width of the box in the x direction in.
+        Default is 1.
+    height: int, optional
+        Height of the box in the y direction.
+        Default is 1.
+    depth: int, optional
+        Depth of the box in the z direction.
+        Default is 1.
 
     Returns
     -------
     array
         2x3 array of lower and upper bounds (in that order).
     """
-    xl = ctr - np.array([delx, dely, delz])
-    xh = ctr + np.array([delx, dely, delz])
+    xl = ctr - np.array([width//2, height//2, depth//2])
+    xh = xl + np.array([width, height, depth])
     return np.array([xl, xh])
 
 
@@ -84,7 +77,7 @@ class ImageryClient(object):
         Datastack name to lookup information for in the InfoService, by default None
     server_address : str, optional
         Address of an InfoService host, by default None. If none, uses defaults in
-        the annotationframeworkclient.
+        the CAVEclient.
     base_resolution : array-like or 'image' or 'segmentation', optional
         Sets the voxel resolution that bounds will be entered in, by default 'image'.
         Literal resolutions will be followed, while "image" or "segmentation" use the
@@ -100,10 +93,10 @@ class ImageryClient(object):
         If False, no segmentation cloudvolume is initialized. By default True
     imagery : bool, optional
         If False, no imagery cloudvolume is initialized. By default True
-    framework_client : annotationframeworkclient.FrameworkClient, optional
+    framework_client : caveclient.CAVEclient, optional
         A pre-initialized Framework client to be used instead of initializing a new one.
     auth_token : str or None, optional
-        Auth token to use for cloudvolume. If None, uses the default values from the FrameworkClient. Default is None.
+        Auth token to use for cloudvolume. If None, uses the default values from the CAVEclient. Default is None.
     timestamp : datetime.datetime or None,
         Fixed timestamp to use for segmentation lookups. If None, defaults to the present time
         when each function is run. Default is None.
@@ -129,14 +122,19 @@ class ImageryClient(object):
         if framework_client is not None:
             self._client = framework_client
         elif datastack_name is not None:
-            self._client = FrameworkClient(
+            try:
+                from caveclient import CAVEclient
+            except:
+                raise ImportError(
+                    'You need to install caveclient to use this functionality')
+            self._client = CAVEclient(
                 datastack_name, server_address=server_address)
         else:
             self._client = None
 
-        if self.client is not None:
-            auth_token = self.client.auth.token
-        _patch_cloudvolume_auth(auth_token)
+        self._auth_token = None
+        if auth_token is not None:
+            self._auth_token = auth_token
 
         if isinstance(base_resolution, str):
             if base_resolution in ['image', 'segmentation']:
@@ -160,6 +158,15 @@ class ImageryClient(object):
     def client(self):
         return self._client
 
+    @property
+    def token(self):
+        if self._auth_token is not None:
+            return self._auth_token
+        elif self.client is None:
+            return None
+        else:
+            return self.client.auth.token
+
     @property
     def timestamp(self):
         if self._timestamp is None:
@@ -212,8 +219,10 @@ class ImageryClient(object):
             self._img_cv = cv.CloudVolume(self.image_source,
                                           mip=self._base_imagery_mip,
                                           bounded=False,
+                                          fill_missing=True,
                                           progress=False,
-                                          use_https=True)
+                                          use_https=True,
+                                          secrets=self.token)
         return self._img_cv
 
     @property
@@ -227,8 +236,10 @@ class ImageryClient(object):
             self._seg_cv = cv.CloudVolume(self.segmentation_source,
                                           mip=self._base_segmentation_mip,
                                           use_https=True,
+                                          fill_missing=True,
                                           bounded=False,
-                                          progress=False)
+                                          progress=False,
+                                          secrets=self.token)
         return self._seg_cv
 
     def _scale_nm_to_voxel(self, xyz_nm, mip, use_cv):
@@ -248,20 +259,33 @@ class ImageryClient(object):
         return bounds_mip_voxel
 
     def _bounds_to_slices(self, bounds):
-        lbounds = np.min(bounds, axis=0)
-        ubounds = np.max(bounds, axis=0)+1
-        xslice, yslice, zslice = (slice(lbounds[ii], ubounds[ii]) if ubounds[ii] > lbounds[ii]+1
-                                  else lbounds[ii] for ii in range(3))
-        return xslice, yslice, zslice
+        return cv.Bbox(bounds[0], bounds[1])
 
-    def image_cutout(self, bounds, mip=None):
+    def _compute_bounds(self, bounds, mip, use_cv, voxel_dimensions):
+        bounds_vx = self._rescale_for_mip(bounds, mip, use_cv=use_cv)
+        if voxel_dimensions is not None:
+            if len(voxel_dimensions) == 3:
+                width, height, depth = voxel_dimensions
+            elif len(voxel_dimensions) == 2:
+                width, height = voxel_dimensions
+                depth = 1
+            bounds_vx = bounds_from_center(np.atleast_2d(bounds_vx)[0],
+                                           width=width, height=height, depth=depth)
+        bbox = self._bounds_to_slices(bounds_vx)
+        return bbox
+
+    def image_cutout(self, bounds, voxel_dimensions=None, mip=None):
         """Get an image cutout for a certain location or set of bounds and a mip level.
 
         Parameters
         ----------
-        bounds : 2 x 3 list of ints
-            A list of a lower bound and upper bound point to bound the cutout in units of voxels in a resolution set by
-            the base_resolution parameter
+        bounds : array
+            Either a 2x3 array of a lower bound and upper bound point to bound the cutout in units of voxels in a resolution set by
+            the base_resolution parameter. Alternatively, if voxel_dimensions is set, bounds is a 3-element array specifying the center of the
+            field of view.
+        voxel_dimensons : array or None, optional
+            If not None, voxel_dimensons is a 3 element array of ints giving the dimensions. In this case, bounds is treated
+            as the center.
         mip : int, optional
             Mip level of imagery to get if something other than the default is wanted, by default None
 
@@ -275,11 +299,10 @@ class ImageryClient(object):
 
         if mip is None:
             mip = self._base_imagery_mip
-        bounds_vx = self._rescale_for_mip(bounds, mip, use_cv='image')
-        slices = self._bounds_to_slices(bounds_vx)
-        return np.array(np.squeeze(self.image_cv.download(slices, mip=mip)))
+        bbox = self._compute_bounds(bounds, mip, 'image', voxel_dimensions)
+        return np.array(np.squeeze(self.image_cv.download(bbox, mip=mip)))
 
-    def segmentation_cutout(self, bounds, root_ids='all', mip=None, verbose=False):
+    def segmentation_cutout(self, bounds, root_ids='all', mip=None, voxel_dimensions=None,  verbose=False):
         """Get a cutout of the segmentation imagery for some or all root ids between set bounds.
         Note that if all root ids are requested in a large region, it could take a long time to query
         all supervoxels.
@@ -294,12 +317,13 @@ class ImageryClient(object):
             find all root ids corresponding to the supervoxels in the cutout and get all of them. None, by default 'all'
         mip : int, optional
             Mip level of the segmentation if something other than the defualt is wanted, by default None
-        verbose : bool, optional
-            If true, prints statements about the progress as it goes. By default, False.
+        voxel_dimensons : array or None, optional
+            If not None, voxel_dimensons is a 3 element array of ints giving the dimensions. In this case, bounds is treated
+            as the center.
 
         Returns
         -------
-        numpy.ndarray 
+        numpy.ndarray
             Array whose elements correspond to the root id (or, if root_ids=None, the supervoxel id) at each voxel.
         """
         if self.segmentation_cv is None:
@@ -308,19 +332,20 @@ class ImageryClient(object):
         if mip is None:
             mip = self._base_segmentation_mip
 
-        bounds_vx = self._rescale_for_mip(bounds, mip, use_cv='segmentation')
-        slices = self._bounds_to_slices(bounds_vx)
+        bbox = self._compute_bounds(
+            bounds, mip, 'segmentation', voxel_dimensions)
+
         if isinstance(root_ids, str):
             if root_ids == 'all':
                 seg_cutout = self.segmentation_cv.download(
-                    slices, agglomerate=True, mip=mip, timestamp=self.timestamp)
+                    bbox, agglomerate=True, mip=mip, timestamp=self.timestamp)
                 return np.array(np.squeeze(seg_cutout))
             else:
                 raise ValueError('root_ids must be None, list, or "all')
         else:
-            return np.array(np.squeeze(self.segmentation_cv.download(slices, segids=root_ids, mip=mip, timestamp=self.timestamp)))
+            return np.array(np.squeeze(self.segmentation_cv.download(bbox, segids=root_ids, mip=mip, timestamp=self.timestamp)))
 
-    def split_segmentation_cutout(self, bounds, root_ids='all', mip=None, include_null_root=False, verbose=False):
+    def split_segmentation_cutout(self, bounds, root_ids='all', mip=None, include_null_root=False, voxel_dimensions=None, verbose=False):
         """Generate segmentation cutouts with a single binary mask for each root id, organized as a dict with keys as root ids and masks as values.
 
         Parameters
@@ -335,6 +360,9 @@ class ImageryClient(object):
             Mip level of the segmentation if something other than the default is wanted, by default None
         include_null_root : bool, optional
             If True, includes root id of 0, which is usually reserved for a null segmentation value. Default is False.
+        voxel_dimensons : array or None, optional
+            If not None, voxel_dimensons is a 3 element array of ints giving the dimensions. In this case, bounds is treated
+            as the center.
         verbose : bool, optional
             If true, prints statements about the progress as it goes. By default, False.
 
@@ -344,7 +372,7 @@ class ImageryClient(object):
             Dict whose keys are root ids and whose values are the binary mask for that root id, with a 1 where the object contains the voxel.
         """
         seg_img = self.segmentation_cutout(
-            bounds=bounds, root_ids=root_ids, mip=mip, verbose=verbose)
+            bounds=bounds, root_ids=root_ids, mip=mip, voxel_dimensions=voxel_dimensions, verbose=verbose)
         return self.segmentation_masks(seg_img, include_null_root)
 
     def segmentation_masks(self, seg_img, include_null_root=False):
@@ -370,7 +398,8 @@ class ImageryClient(object):
             split_segmentation[root_id] = (seg_img == root_id).astype(int)
         return split_segmentation
 
-    def image_and_segmentation_cutout(self, bounds, image_mip=None, segmentation_mip=None, root_ids='all', resize=True, split_segmentations=False, include_null_root=False, verbose=False):
+    def image_and_segmentation_cutout(self, bounds, image_mip=None, segmentation_mip=None, root_ids='all', resize=True,
+                                      split_segmentations=False, include_null_root=False, voxel_dimensions=None, verbose=False):
         """Download aligned and scaled imagery and segmentation data at a given resolution.
 
         Parameters
@@ -393,12 +422,15 @@ class ImageryClient(object):
             an array with root_ids (using segmentation_cutout), by default False
         include_null_root : bool, optional
             If True, includes root id of 0, which is usually reserved for a null segmentation value. Default is False.
+        voxel_dimensons : array or None, optional
+            If not None, voxel_dimensons is a 3 element array of ints giving the dimensions of the cutout. In this case, bounds is treated
+            as the center.
         verbose : bool, optional
             If True, prints statements about the progress as it goes. By default, False.
 
         Returns
         -------
-        cloudvolume.VolumeCutout 
+        cloudvolume.VolumeCutout
             Imagery volume cutout
 
         numpy.ndarray or dict
@@ -413,15 +445,38 @@ class ImageryClient(object):
         seg_resolution = self.segmentation_cv.mip_resolution(segmentation_mip)
         if np.all(img_resolution == seg_resolution):
             zoom_to = None
-        if np.all(img_resolution >= seg_resolution):
+            if voxel_dimensions is not None:
+                image_voxel_dimensions = voxel_dimensions
+                segmentation_voxel_dimensions = voxel_dimensions
+            else:
+                image_voxel_dimensions = None
+                segmentation_voxel_dimensions = None
+        elif np.all(img_resolution >= seg_resolution):
             zoom_to = 'segmentation'
+            res_scaling = np.array(seg_resolution) / np.array(img_resolution)
+            if voxel_dimensions is not None:
+                segmentation_voxel_dimensions = voxel_dimensions
+                image_voxel_dimensions = [int(vd * rs) for vd,
+                                          rs in zip(segmentation_voxel_dimensions, res_scaling)]
+            else:
+                image_voxel_dimensions = None
+                segmentation_voxel_dimensions = None
         else:
             zoom_to = 'image'
+            res_scaling = np.array(img_resolution) / np.array(seg_resolution)
+            if voxel_dimensions is not None:
+                image_voxel_dimensions = voxel_dimensions
+                segmentation_voxel_dimensions = [int(vd * rs) for vd,
+                                                 rs in zip(image_voxel_dimensions, res_scaling)]
+            else:
+                image_voxel_dimensions = None
+                segmentation_voxel_dimensions = None
 
         if verbose:
             print('Downloading images')
         img = self.image_cutout(bounds=bounds,
-                                mip=image_mip)
+                                mip=image_mip,
+                                voxel_dimensions=image_voxel_dimensions)
         img_shape = img.shape
 
         if verbose:
@@ -430,6 +485,7 @@ class ImageryClient(object):
             seg = self.segmentation_cutout(bounds,
                                            root_ids=root_ids,
                                            mip=segmentation_mip,
+                                           voxel_dimensions=segmentation_voxel_dimensions,
                                            verbose=verbose)
             seg_shape = seg.shape
         else:
@@ -437,6 +493,7 @@ class ImageryClient(object):
                                                  root_ids=root_ids,
                                                  mip=segmentation_mip,
                                                  include_null_root=include_null_root,
+                                                 voxel_dimensions=segmentation_voxel_dimensions,
                                                  verbose=verbose)
             if len(seg) > 0:
                 seg_shape = seg[list(seg.keys())[0]].shape
@@ -514,7 +571,7 @@ class ImageryClient(object):
                 self.segmentation_cv.mip_resolution(mip))] = mip
         return image_resolution, segmentation_resolution
 
-    def save_imagery(self, filename_prefix, bounds=None, mip=None, precomputed_image=None, slice_axis=2, verbose=False, **kwargs):
+    def save_imagery(self, filename_prefix, bounds=None, mip=None, precomputed_image=None, slice_axis=2, voxel_dimensions=None, verbose=False, **kwargs):
         """Save queried or precomputed imagery to png files.
 
         Parameters
@@ -536,14 +593,16 @@ class ImageryClient(object):
             If True, prints the progress, by default False
         """
         if precomputed_image is None:
-            img = self.image_cutout(bounds, mip)
+            img = self.image_cutout(
+                bounds, mip=mip, voxel_dimensions=voxel_dimensions)
         else:
             img = precomputed_image
-        utils._save_image_slices(filename_prefix, 'imagery', img, slice_axis,
-                                 'imagery', verbose=verbose, **kwargs)
-        return
+        save_image_slices(filename_prefix, 'imagery', img, slice_axis,
+                          'imagery', verbose=verbose,  **kwargs)
+        pass
 
-    def save_segmentation_masks(self, filename_prefix, bounds=None, mip=None, root_ids='all', precomputed_masks=None, slice_axis=2, include_null_root=False, segmentation_colormap={}, verbose=False, **kwargs):
+    def save_segmentation_masks(self, filename_prefix, bounds=None, mip=None, root_ids='all', precomputed_masks=None,
+                                slice_axis=2, include_null_root=False, segmentation_colormap={}, verbose=False, voxel_dimensions=None, **kwargs):
         """Save queried or precomputed segmentation masks to png files. Additional kwargs are passed to imageio.imwrite.
 
         Parameters
@@ -576,18 +635,18 @@ class ImageryClient(object):
         """
         if precomputed_masks is None:
             seg_dict = self.split_segmentation_cutout(
-                bounds=bounds, root_ids=root_ids, mip=mip, include_null_root=include_null_root)
+                bounds=bounds, root_ids=root_ids, mip=mip, voxel_dimensions=voxel_dimensions, include_null_root=include_null_root)
         else:
             seg_dict = precomputed_masks
 
         for root_id, seg_mask in seg_dict.items():
             suffix = f'root_id_{root_id}'
-            utils._save_image_slices(filename_prefix, suffix, seg_mask, slice_axis, 'mask',
-                                     color=segmentation_colormap.get(root_id, None), verbose=verbose, **kwargs)
-        return
+            save_image_slices(filename_prefix, suffix, seg_mask, slice_axis, 'mask',
+                              color=segmentation_colormap.get(root_id, None), verbose=verbose, **kwargs)
+        pass
 
     def save_image_and_segmentation_masks(self, filename_prefix, bounds=None, image_mip=None, segmentation_mip=None,
-                                          root_ids='all', resize=True, precomputed_data=None, slice_axis=2,
+                                          root_ids='all', resize=True, precomputed_data=None, slice_axis=2, voxel_dimensions=None,
                                           segmentation_colormap={}, include_null_root=False, verbose=False, **kwargs):
         """Save aligned and scaled imagery and segmentation mask cutouts as pngs. Kwargs are passed to imageio.imwrite.
 
@@ -627,11 +686,11 @@ class ImageryClient(object):
         if precomputed_data is not None:
             img, seg_dict = precomputed_data
         else:
-            img, seg_dict = self.image_and_segmentation_cutout(bounds=bounds, image_mip=image_mip, segmentation_mip=segmentation_mip,
+            img, seg_dict = self.image_and_segmentation_cutout(bounds=bounds, image_mip=image_mip, segmentation_mip=segmentation_mip, voxel_dimensions=voxel_dimensions,
                                                                root_ids=root_ids, resize=resize, include_null_root=include_null_root, split_segmentations=True, verbose=verbose)
 
         self.save_imagery(filename_prefix, precomputed_image=img,
                           slice_axis=slice_axis, verbose=verbose, **kwargs)
         self.save_segmentation_masks(filename_prefix, precomputed_masks=seg_dict, slice_axis=slice_axis,
                                      segmentation_colormap=segmentation_colormap, verbose=verbose, **kwargs)
-        return
+        pass
